"""Export RT-DC measurement data"""
from __future__ import annotations

import codecs
import json
import pathlib
import time
from typing import Dict, List
import uuid
import warnings

import h5py
import hdf5plugin

try:
    import imageio
except ModuleNotFoundError:
    IMAGEIO_AVAILABLE = False
else:
    IMAGEIO_AVAILABLE = True

try:
    import fcswrite
except ModuleNotFoundError:
    FCSWRITE_AVAILABLE = False
else:
    FCSWRITE_AVAILABLE = True

import numpy as np

from .. import definitions as dfn
from .._version import version, version_tuple

from .feat_basin import get_basin_classes
from .writer import RTDCWriter


class LimitingExportSizeWarning(UserWarning):
    pass


class Export(object):
    def __init__(self, rtdc_ds):
        """Export functionalities for RT-DC datasets"""
        self.rtdc_ds = rtdc_ds

    def avi(self, path, filtered=True, override=False):
        """Exports filtered event images to an avi file

        Parameters
        ----------
        path: str
            Path to a .avi file. The ending .avi is added automatically.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.

        Notes
        -----
        Raises OSError if current dataset does not contain image data
        """
        if not IMAGEIO_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `imageio` required for avi export!")
        path = pathlib.Path(path)
        ds = self.rtdc_ds
        # Make sure that path ends with .avi
        if path.suffix != ".avi":
            path = path.with_name(path.name + ".avi")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Start exporting
        if "image" in ds:
            # Open video for writing
            vout = imageio.get_writer(uri=path,
                                      format="FFMPEG",
                                      fps=25,
                                      codec="rawvideo",
                                      pixelformat="yuv420p",
                                      macro_block_size=None,
                                      ffmpeg_log_level="error")
            # write the filtered frames to avi file
            for evid in np.arange(len(ds)):
                # skip frames that were filtered out
                if filtered and not ds.filter.all[evid]:
                    continue
                image = ds["image"][evid]
                # Convert image to RGB
                image = image.reshape(image.shape[0], image.shape[1], 1)
                image = np.repeat(image, 3, axis=2)
                vout.append_data(image)
        else:
            msg = "No image data to export: dataset {} !".format(ds.title)
            raise OSError(msg)

    def fcs(self, path, features, meta_data=None, filtered=True,
            override=False):
        """Export the data of an RT-DC dataset to an .fcs file

        Parameters
        ----------
        path: str
            Path to an .fcs file. The ending .fcs is added automatically.
        features: list of str
            The features in the resulting .fcs file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            in the primary TEXT segment of the FCS file; the version
            of dclab is stored there by default
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.

        Notes
        -----
        Due to incompatibility with the .fcs file format, all events with
        NaN-valued features are not exported.
        """
        if meta_data is None:
            meta_data = {}
        if not FCSWRITE_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `fcswrite` required for fcs export!")

        ds = self.rtdc_ds

        path = pathlib.Path(path)
        # Make sure that path ends with .fcs
        if path.suffix != ".fcs":
            path = path.with_name(path.name + ".fcs")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features are valid
        features = sorted(set(features))
        for c in features:
            if c not in ds.features_scalar:
                msg = "Invalid feature name: {}".format(c)
                raise ValueError(msg)

        # Collect the header
        chn_names = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]

        # Collect the data
        if filtered:
            data = [ds[c][ds.filter.all] for c in features]
        else:
            data = [ds[c] for c in features]

        data = np.array(data).transpose()
        meta_data["dclab version"] = version
        fcswrite.write_fcs(filename=str(path),
                           chn_names=chn_names,
                           data=data,
                           text_kw_pr=meta_data,
                           )

    def hdf5(self,
             path: str | pathlib.Path,
             features: List[str] = None,
             filtered: bool = True,
             logs: bool = False,
             tables: bool = False,
             basins: bool = False,
             meta_prefix: str = "src_",
             override: bool = False,
             compression_kwargs: Dict = None,
             compression: str = "deprecated",
             skip_checks: bool = False):
        """Export the data of the current instance to an HDF5 file

        Parameters
        ----------
        path: str
            Path to an .rtdc file. The ending .rtdc is added
            automatically.
        features: list of str
            The features in the resulting .rtdc file. These are strings
            that are defined by `dclab.definitions.feature_exists`, e.g.
            "area_cvx", "deform", "frame", "fl1_max", "image".
            Defaults to `self.rtdc_ds.features_innate`.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        logs: bool
            Whether to store the logs of the original file prefixed with
            `source_` to the output file.
        tables: bool
            Whether to store the tables of the original file prefixed with
            `source_` to the output file.
        basins: bool
            Whether to export basins. If filtering is disabled, basins
            are copied directly to the output file. If filtering is enabled,
            then mapped basins are exported.
        meta_prefix: str
            Prefix for log and table names in the exported file
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        compression_kwargs: dict
            Dictionary with the keys "compression" and "compression_opts"
            which are passed to :func:`h5py.H5File.create_dataset`. The
            default is Zstandard compression with the lowest compression
            level `hdf5plugin.Zstd(clevel=1)`.
        compression: str or None
            Compression method used for data storage;
            one of [None, "lzf", "gzip", "szip"].

            .. deprecated:: 0.43.0
                Use `compression_kwargs` instead.
        skip_checks: bool
            Disable checking whether all features have the same length.


        .. versionchanged:: 0.58.0

           The ``basins`` keyword argument was added, and it is now possible
           to pass an empty list to ``features``. This combination results
           in a very small file consisting of metadata and a mapped basin
           referring to the original dataset.
        """
        if compression != "deprecated":
            warnings.warn("The `compression` kwarg is deprecated in favor of "
                          "`compression_kwargs`!",
                          DeprecationWarning)
            if compression_kwargs is not None:
                raise ValueError("You may not specify `compression` and "
                                 "`compression_kwargs` at the same time!")
            # be backwards-compatible
            compression_kwargs = {"compression": compression}
        if compression_kwargs is None:
            compression_kwargs = hdf5plugin.Zstd(clevel=1)
        path = pathlib.Path(path)
        # Make sure that path ends with .rtdc
        if path.suffix not in [".rtdc", ".rtdc~"]:
            path = path.parent / (path.name + ".rtdc")
        # Check if file already exists
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(path)
                          + "Please use the `override=True` option.")
        elif path.exists():
            path.unlink()

        # make sure the parent directory exists
        path.parent.mkdir(parents=True, exist_ok=True)

        # for convenience
        ds = self.rtdc_ds

        if features is None:
            features = ds.features_innate

        # decide which metadata to export
        meta = {}
        # only cfg metadata (no analysis metadata)
        for sec in dfn.CFG_METADATA:
            if sec in ds.config:
                meta[sec] = ds.config[sec].copy()
        # add user-defined metadata
        if "user" in ds.config:
            meta["user"] = ds.config["user"].copy()
        if filtered:
            # Define a new measurement identifier, so that we are not running
            # into any problems with basins being defined for filtered data.
            ds_run_id = ds.get_measurement_identifier()
            random_ap = str(uuid.uuid4())[:4]
            meta["experiment"]["run identifier"] = f"{ds_run_id}-{random_ap}"

        if filtered:
            filter_arr = ds.filter.all
        else:
            filter_arr = None

        features = sorted(set(features))
        if not skip_checks and features:
            # check that all features have same length and use the smallest
            # common length
            lengths = []
            for feat in features:
                if feat == "trace":
                    for tr in list(ds["trace"].keys()):
                        lengths.append(len(ds["trace"][tr]))
                else:
                    lengths.append(len(ds[feat]))
            l_min = np.min(lengths)
            l_max = np.max(lengths)
            if l_min != l_max:
                if filter_arr is None:
                    # we are forced to do filtering
                    filter_arr = np.ones(len(ds), dtype=bool)
                else:
                    # have to create a copy, because rtdc_ds.filter.all is ro!
                    filter_arr = np.copy(filter_arr)
                filter_arr[l_min:] = False
                warnings.warn(
                    "Not all features have the same length! Limiting output "
                    + f"event count to {l_min} (max {l_max}) in '{l_min}'.",
                    LimitingExportSizeWarning)

        # Perform actual export
        with RTDCWriter(path,
                        mode="append",
                        compression_kwargs=compression_kwargs) as hw:
            # write meta data
            hw.store_metadata(meta)

            # write export log
            hw.store_log(time.strftime("dclab-export_%Y-%m-%d_%H.%M.%S"),
                         json.dumps(
                             {"dclab version": version_tuple,
                              "kwargs": {
                                  "features": features,
                                  "filtered": filtered,
                                  "logs": logs,
                                  "tables": tables,
                                  "basins": basins,
                                  "meta_prefix": meta_prefix,
                                  "skip_checks": skip_checks
                                  }
                              },
                             indent=2,
                             sort_keys=True,
                         ).split("\n"))

            if logs:
                # write logs
                for log in ds.logs:
                    hw.store_log(f"{meta_prefix}{log}",
                                 ds.logs[log])

            if tables:
                # write tables
                for tab in ds.tables:
                    hw.store_table(f"{meta_prefix}{tab}",
                                   ds.tables[tab])

            # write each feature individually
            for feat in features:
                if (filter_arr is None or
                        # This does not work for the .tdms file format
                        # (and probably also not for DCOR).
                        (np.all(filter_arr) and ds.format == "hdf5")):
                    # We do not have to filter and can be fast
                    if dfn.scalar_feature_exists(feat):
                        shape = (1,)
                    elif feat in ["image", "image_bg", "mask", "trace"]:
                        # known shape
                        shape = None
                    else:
                        shape = np.array(ds[feat][0]).shape
                    hw.store_feature(feat=feat,
                                     data=ds[feat],
                                     shape=shape)
                else:
                    # We have to filter and will be slower
                    store_filtered_feature(rtdc_writer=hw,
                                           feat=feat,
                                           data=ds[feat],
                                           filtarr=filter_arr)

            if basins:
                # We have to store basins. There are three options:
                # - filtering disabled: just copy basins
                # - filtering enabled
                #   - basins with "same" mapping: create new mapping
                #   - mapped basins: correct nested mapping
                # In addition to the basins that we copy from the
                # original dataset, we also create a new basin that
                # refers to the original dataset itself.
                basin_list = [bn.as_dict() for bn in ds.basins]
                # In addition to the upstream basins, also store a reference
                # to the original file from which the export was done.
                if ds.format in get_basin_classes():
                    # The dataset has a format that matches a basin format
                    # directly.
                    basin_is_local = ds.format == "hdf5"
                    basin_locs = [ds.path]
                    if basin_is_local:
                        # So the user can put them into the same directory.
                        basin_locs.append(ds.path.name)
                    basin_list.append({
                        "basin_name": "Exported data",
                        "basin_type": "file" if basin_is_local else "remote",
                        "basin_format": ds.format,
                        "basin_locs": basin_locs,
                        "basin_descr": f"Exported with dclab {version}",
                    })
                elif (ds.format == "hierarchy"
                      and ds.get_root_parent().format in get_basin_classes()):
                    # avoid circular imports
                    from .fmt_hierarchy import map_indices_child2root
                    # The dataset is a hierarchy child, and it is derived
                    # from a dataset that has a matching basin format.
                    # We have to add the indices of the root parent, which
                    # identify the child, to the basin dictionary. Note
                    # that additional basin filtering is applied below
                    # this case for all basins.
                    # For the sake of clarity I wrote this as a separate case,
                    # even if that means duplicating code from the previous
                    # case.
                    ds_root = ds.get_root_parent()
                    basin_is_local = ds_root.format == "hdf5"
                    basin_locs = [ds_root.path]
                    if basin_is_local:
                        # So the user can put them into the same directory.
                        basin_locs.append(ds_root.path.name)
                    basin_list.append({
                        "basin_name": "Exported data (hierarchy)",
                        "basin_type": "file" if basin_is_local else "remote",
                        "basin_format": ds_root.format,
                        "basin_locs": basin_locs,
                        "basin_descr": f"Exported with dclab {version} from a "
                                       f"hierarchy dataset",
                        # This is where this basin differs from the basin
                        # definition in the previous case.
                        "basin_map": map_indices_child2root(
                            child=ds,
                            child_indices=np.arange(len(ds))
                            ),
                    })

                for bn_dict in basin_list:
                    if bn_dict.get("basin_type") == "internal":
                        # Internal basins are only valid for files they were
                        # defined in. Since we are exporting, it does not
                        # make sense to store these basins in the output file.
                        continue
                    basinmap_orig = bn_dict.get("basin_map")
                    if not filtered:
                        # filtering disabled: just copy basins
                        pass
                    elif basinmap_orig is None:
                        # basins with "same" mapping: create new mapping
                        bn_dict["basin_map"] = np.where(filter_arr)[0]
                    else:
                        # mapped basins: correct nested mapping
                        bn_dict["basin_map"] = basinmap_orig[filter_arr]

                    # Do not verify basins, it takes too long.
                    hw.store_basin(**bn_dict, verify=False)

    def tsv(self, path, features, meta_data=None, filtered=True,
            override=False):
        """Export the data of the current instance to a .tsv file

        Parameters
        ----------
        path: str
            Path to a .tsv file. The ending .tsv is added automatically.
        features: list of str
            The features in the resulting .tsv file. These are strings
            that are defined by `dclab.definitions.scalar_feature_exists`,
            e.g. "area_cvx", "deform", "frame", "fl1_max", "aspect".
        meta_data: dict
            User-defined, optional key-value pairs that are stored
            at the beginning of the tsv file - one key-value pair is
            stored per line which starts with a hash. The version of
            dclab is stored there by default.
        filtered: bool
            If set to `True`, only the filtered data
            (index in ds.filter.all) are used.
        override: bool
            If set to `True`, an existing file ``path`` will be overridden.
            If set to `False`, raises `OSError` if ``path`` exists.
        """
        if meta_data is None:
            meta_data = {}
        features = [c.lower() for c in features]
        features = sorted(set(features))
        path = pathlib.Path(path)
        ds = self.rtdc_ds
        # Make sure that path ends with .tsv
        if path.suffix != ".tsv":
            path = path.with_name(path.name + ".tsv")
        # Check if file already exist
        if not override and path.exists():
            raise OSError("File already exists: {}\n".format(
                str(path).encode("ascii", "ignore")) +
                "Please use the `override=True` option.")
        # Check that features exist
        for c in features:
            if c not in ds.features_scalar:
                raise ValueError("Invalid feature name {}".format(c))
        meta_data["dclab version"] = version
        # Write BOM header
        with path.open("wb") as fd:
            fd.write(codecs.BOM_UTF8)
        # Open file
        with path.open("a", encoding="utf-8") as fd:
            # write meta data
            for key in sorted(meta_data.keys()):
                fd.write(f"# {key}: {meta_data[key]}\n")
            fd.write("#\n")
            fd.write("# Original dataset configuration:\n")
            cfg = self.rtdc_ds.config.as_dict()
            for sec in sorted(cfg.keys()):
                for key in sorted(cfg[sec].keys()):
                    fd.write(f"# dc:{sec}:{key} = {cfg[sec][key]}\n")
            fd.write("#\n")
            # write header
            header1 = "\t".join([c for c in features])
            fd.write("# "+header1+"\n")
            labels = [dfn.get_feature_label(c, rtdc_ds=ds) for c in features]
            header2 = "\t".join(labels)
            fd.write("# "+header2+"\n")

        with path.open("ab") as fd:
            # write data
            if filtered:
                data = [ds[c][ds.filter.all] for c in features]
            else:
                data = [ds[c] for c in features]

            np.savetxt(fd,
                       np.array(data).transpose(),
                       fmt=str("%.10e"),
                       delimiter="\t")


def yield_filtered_array_stacks(data, indices):
    """Generator returning chunks with the filtered feature data

    Parameters
    ----------
    data: np.ndarray or h5py.Dataset
        The full, unfiltered input feature data. Must implement
        the `shape` and `dtype` properties. If it implements the
        `__array__` method, fast slicing is used.
    indices: np.ndarray or list
        The indices (integer values) for `data` (first axis), indicating
        which elements should be returned by this generator.

    Notes
    -----
    This method works with any feature dimension (e.g. it
    works for image (2D) data and for trace data (1D)). It
    is just important that `data` is indexable using integers
    and that the events in `data` all have the same shape.
    The dtype of the returned chunks is determined by the first
    item in `data`.

    This method works with sliceable (e.g. np.ndarray) and
    non-sliceable (e.g. tdms-format-based images) input data. If the
    input data is sliceable (which is determined by the availability
    of the `__array__` method, then fast numpy sclicing is used. If the
    input data does not support slicing (`__array__` not defined), then
    a slow iteration over `indices` is done.

    In the slow iteration case, the returned array data are overridden
    in-place. If you need to retain a copy of the `yield`ed chunks,
    apply `np.array(.., copy=True)` to the returned chunks.
    """
    chunk_shape = RTDCWriter.get_best_nd_chunks(item_shape=data.shape[1:],
                                                item_dtype=data.dtype)
    chunk_size = chunk_shape[0]

    if hasattr(data, "__array__"):
        # We have an array-like object and can do slicing with the indexing
        # array. This speeds up chunk creation for e.g. the HDF5 file format
        # where all data are present in an array-like fashion.
        indices = np.array(indices)
        stop = 0
        for kk in range(len(indices) // chunk_size):
            start = chunk_size * kk
            stop = chunk_size * (kk + 1)
            yield data[indices[start:stop]]
        if stop < len(indices):
            yield data[indices[stop:]]
    else:
        # assemble filtered image stacks
        chunk = np.zeros(chunk_shape, dtype=data.dtype)
        jj = 0
        for ii in indices:
            chunk[jj] = data[ii]
            if (jj + 1) % chunk_size == 0:
                jj = 0
                yield chunk
            else:
                jj += 1
        # yield remainder
        if jj:
            yield chunk[:jj]


def store_filtered_feature(rtdc_writer, feat, data, filtarr):
    """Append filtered feature data to an HDF5 file

    Parameters
    ----------
    rtdc_writer: dclab.rtdc_dataset.writer.RTDCWriter
        an open writer object
    feat: str
        feature name
    data: object or list or np.ndarray or dict
        feature data
    filtarr: boolean np.ndarray
        filtering array (same as RTDCBase.filter.all)

    Notes
    -----
    This code is somewhat redundant to the code of RTDCWriter.
    """
    indices = np.where(filtarr)[0]
    if indices.size == 0:
        warnings.warn(f"No data to export to '{rtdc_writer.path}'")
        return

    hw = rtdc_writer
    if not hw.mode == "append":
        raise ValueError("The `rtdc_writer` object must be created with"
                         + f"`mode='append'`, got '{hw.mode}' for '{hw}'!")
    # event-wise, because
    # - tdms-based datasets don't allow indexing with numpy
    # - there might be memory issues
    if feat == "contour":
        for ii in indices:
            hw.store_feature("contour", data[ii])
    elif feat in ["mask", "image", "image_bg"]:
        # assemble filtered image stacks
        for imstack in yield_filtered_array_stacks(data, indices):
            hw.store_feature(feat, imstack)
    elif feat == "trace":
        # assemble filtered trace stacks
        for tr in data.keys():
            for trstack in yield_filtered_array_stacks(data[tr], indices):
                hw.store_feature("trace", {tr: trstack})
    elif dfn.scalar_feature_exists(feat):
        hw.store_feature(feat, data[filtarr])
    else:
        # Special case of plugin or temporary features.
        shape = data[0].shape
        for dstack in yield_filtered_array_stacks(data, indices):
            hw.store_feature(feat, dstack, shape=shape)


def hdf5_append(h5obj, rtdc_ds, feat, compression, filtarr=None,
                time_offset=0):
    """Append feature data to an HDF5 file

    Parameters
    ----------
    h5obj: h5py.File
        Opened HDF5 file
    rtdc_ds: dclab.rtdc_dataset.RTDCBase
        Instance from which to obtain the data
    feat: str
        Valid feature name in `rtdc_ds`
    compression: str or None
        Compression method for "contour", "image", and "trace" data
        as well as logs; one of [None, "lzf", "gzip", "szip"].
    filtarr: None or 1d boolean np.ndarray
        Optional boolean array used for filtering. If set to
        `None`, all events are saved.
    time_offset: float
        Do not use! Please use `dclab.cli.task_join.join` instead.

    Notes
    -----
    Please update the "experiment::event count" attribute manually.
    You may use
    :func:`dclab.rtdc_dataset.writer.RTDCWriter.rectify_metadata`
    for that or use the `RTDCWriter` context manager where it is
    automatically run during `__exit__`.
    """
    # optional array for filtering events
    if filtarr is None:
        filtarr = np.ones(len(rtdc_ds), dtype=bool)
        no_filter = True
    else:
        no_filter = False

    warnings.warn("`hdf5_append` is deptecated; please use "
                  " the dclab.RTDCWriter context manager or the "
                  " export.store_filtered_feature function.",
                  DeprecationWarning)

    if time_offset != 0:
        raise ValueError("Setting `time_offset` not supported anymore! "
                         "Please use `dclab.cli.task_join.join` instead.")

    # writer instance
    hw = RTDCWriter(h5obj, mode="append", compression=compression)
    if no_filter:
        hw.store_feature(feat, rtdc_ds[feat])
    else:
        store_filtered_feature(rtdc_writer=hw,
                               feat=feat,
                               data=rtdc_ds[feat],
                               filtarr=filtarr)


def hdf5_autocomplete_config(path_or_h5obj):
    """Autocomplete the configuration of the RTDC-measurement

    The following configuration keys are updated:

    - experiment:event count
    - fluorescence:samples per event
    - imaging: roi size x (if image or mask is given)
    - imaging: roi size y (if image or mask is given)

    The following configuration keys are added if not present:

    - fluorescence:channel count

    Parameters
    ----------
    path_or_h5obj: pathlib.Path or str or h5py.File
        Path to or opened RT-DC measurement
    """
    warnings.warn("`hdf5_autocomplete_config` is deptecated; please use "
                  " the dclab.RTDCWriter context manager or the "
                  " dclab.RTDCWriter.rectify_metadata function.",
                  DeprecationWarning)
    if not isinstance(path_or_h5obj, h5py.File):
        close = True
    else:
        close = False

    hw = RTDCWriter(path_or_h5obj, mode="append")
    hw.rectify_metadata()

    if close:
        path_or_h5obj.close()



class H5Logs:
    def __init__(self, h5):
        self.h5file = h5
        self._cache_keys = None

    def __getitem__(self, key):
        if key in self.keys():
            log = list(self.h5file["logs"][key])
            if isinstance(log[0], bytes):
                log = [li.decode("utf") for li in log]
        else:
            raise KeyError(
                f"File {self.h5file.file.filename} does not have the log "
                f"'{key}'. Available logs are {self.keys()}.")
        return log

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self.keys())

    def keys(self):
        if self._cache_keys is None:
            names = []
            if "logs" in self.h5file:
                for key in self.h5file["logs"]:
                    if self.h5file["logs"][key].size:
                        names.append(key)
            self._cache_keys = names
        return self._cache_keys



import numbers
from os.path import join
import tempfile

import h5py
import numpy as np

import pytest

import dclab
from dclab.rtdc_dataset import RTDCWriter, new_dataset
from dclab.rtdc_dataset.feat_temp import (
    register_temporary_feature, deregister_temporary_feature)

from helper_methods import retrieve_data, DCOR_AVAILABLE, BOTO3_AVAILABLE


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_file():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_test = path.parent / "test.h5"
    # We basically create a file that consists only of the metadata.
    with RTDCWriter(path_test) as hw, new_dataset(path) as dsorig:
        hw.store_basin(basin_name="get-out",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[path],
                       basin_descr="My very first basin-only dataset",
                       verify=True,
                       )
        meta = dsorig.config.as_dict(pop_filtering=True)
        hw.store_metadata(meta)

    # OK, now open the dataset and make sure that it contains all information.
    with new_dataset(path_test) as ds:
        assert len(ds) == 5
        assert ds.config["experiment"]["sample"] == "background image example"
        assert np.allclose(ds["deform"][2],
                           0.029980043,
                           atol=0,
                           rtol=1e-5)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_feature_restriction():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_test = path.parent / "test.h5"
    # We basically create a file that consists only of the metadata.
    with RTDCWriter(path_test) as hw, new_dataset(path) as dsorig:
        hw.store_basin(basin_name="get-out",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[path],
                       basin_descr="My very first basin-only dataset",
                       basin_feats=["deform"],
                       verify=True,
                       )
        meta = dsorig.config.as_dict(pop_filtering=True)
        hw.store_metadata(meta)
        # sanity check
        assert "area_um" in dsorig

    # Make sure *only* derorm is in the basin
    with dclab.new_dataset(path_test) as ds:
        assert "deform" in ds.features
        assert "area_um" not in ds.features
        assert "deform" in ds.features_basin


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_feature_restriction_verify():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_test = path.parent / "test.h5"
    # We basically create a file that consists only of the metadata.
    with RTDCWriter(path_test) as hw:
        with pytest.raises(ValueError, match="blablabla"):
            hw.store_basin(basin_name="get-out",
                           basin_type="file",
                           basin_format="hdf5",
                           basin_locs=[path],
                           basin_descr="My very first basin-only dataset",
                           basin_feats=["blablabla"],
                           verify=True,
                           )


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_file_relative():
    """Test whether storing the relative path works"""
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_test = path.parent / "test.h5"
    # We basically create a file that consists only of the metadata.
    with RTDCWriter(path_test) as hw, new_dataset(path) as dsorig:
        hw.store_basin(basin_name="get-out",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[path],
                       basin_descr="My very first basin-only dataset",
                       verify=True
                       )
        meta = dsorig.config.as_dict(pop_filtering=True)
        hw.store_metadata(meta)

    # Now to the relative testing part. We just move everything one directory
    # down. So the absolute paths will not be valid anymore.
    newdir = path.parent / "new"
    newdir.mkdir()
    new_path = newdir / path.name
    new_path_test = newdir / path_test.name
    path.rename(new_path)
    path_test.rename(new_path_test)

    # OK, now open the dataset and make sure that it contains all information.
    with new_dataset(new_path_test) as ds:
        assert len(ds.basins) == 1
        assert len(ds) == 5
        assert ds.config["experiment"]["sample"] == "background image example"
        assert np.allclose(ds["deform"][2],
                           0.029980043,
                           atol=0,
                           rtol=1e-5)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_file_identifier_mismatch():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with RTDCWriter(path) as ho:
        ho.store_metadata({"experiment": {"run identifier": "captain-proton"}})

    path_test = path.parent / "test.h5"
    # We basically create a file that consists only of the metadata.
    with RTDCWriter(path_test) as hw, new_dataset(path) as dsorig:
        meta = dsorig.config.as_dict(pop_filtering=True)
        hw.store_metadata(meta)
        hw.store_metadata({"experiment": {"run identifier": "janeway"}})
        with pytest.raises(ValueError, match="identifier mismatch"):
            hw.store_basin(basin_name="get-out",
                           basin_type="file",
                           basin_format="hdf5",
                           basin_locs=[path],
                           basin_descr="My very first basin-only dataset",
                           verify=True
                           )
        # But it should work when verify is False
        hw.store_basin(basin_name="get-out",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[path],
                       basin_descr="My very first basin-only dataset",
                       verify=False
                       )

    # Then, we have zero basins...
    with new_dataset(path_test) as ds:
        assert len(ds.basins) == 0

    # ...unless we change the run identifier to match the basin.
    with RTDCWriter(path_test) as hw:
        hw.store_metadata({"experiment": {"run identifier": "captain-proton"}})

    with new_dataset(path_test) as ds:
        assert len(ds.basins) == 1


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not reachable!")
@pytest.mark.skipif(not BOTO3_AVAILABLE, reason="boto3 not available!")
def test_basin_url(tmp_path):
    # Create an empty dataset that links to an S3 instance
    # This is the calibration beads measurement.
    # https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2/
    # resource/fb719fb2-bd9f-817a-7d70-f4002af916f0
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")
    with new_dataset(s3_url) as ds:
        meta = ds.config.as_dict(pop_filtering=True)

    path = tmp_path / "basins3test.rtdc"
    with RTDCWriter(path) as hw:
        hw.store_metadata(meta)
        hw.store_basin(basin_name="s3test",
                       basin_type="remote",
                       basin_format="s3",
                       basin_locs=[s3_url],
                       basin_descr="Me very first online dataset",
                       verify=True,
                       )

    with new_dataset(path) as ds:
        assert ds.basins[0].name == "s3test"
        assert ds.basins[0].basin_type == "remote"
        assert ds.basins[0].basin_format == "s3"
        assert ds.basins[0].location == s3_url
        assert ds.basins[0].description == "Me very first online dataset"
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][0],
                           0.009741939,
                           atol=0,
                           rtol=1e-5)


def test_bulk_scalar():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_scalar_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "area_um" in events.keys()
        assert np.all(events["area_um"][:] == data["area_um"])


def test_bulk_contour():
    num = 7
    contour = []
    for ii in range(5, num + 5):
        cii = np.arange(2 * ii).reshape(2, ii)
        contour.append(cii)
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "contour": contour}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_contour_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "contour" in events.keys()
        assert np.allclose(events["contour"]["6"], contour[6])
        assert events["contour"]["1"].shape == (2, 6)


def test_bulk_image():
    num = 7
    image = np.zeros((20, 90, 50))
    image += np.arange(90).reshape(1, 90, 1)
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "image": image}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_image_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "image" in events.keys()
        assert np.allclose(events["image"][6], image[6])


@pytest.mark.parametrize(
    "feat,dtype", [
        ("qpi_oah", np.uint8),
        ("qpi_oah_bg", np.uint8),
        ("qpi_pha", np.float32),
        ("qpi_amp", np.float32),
    ]
)
def test_bulk_qpi2d(feat, dtype):
    num = 7
    feat_data = np.zeros((20, 90, 50), dtype=dtype)
    if dtype == np.uint8:
        feat_data += np.arange(90, dtype=dtype).reshape(1, 90, 1)
    else:
        feat_data += np.linspace(0.8, 1.1, 90, dtype=dtype).reshape(1, 90, 1)
    data = {"area_um": np.linspace(100.7, 110.9, num),
            feat: feat_data}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_qpi_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert feat in events.keys()
        assert np.allclose(events[feat][6], feat_data[6])
        assert events[feat][:].dtype == dtype


def test_bulk_mask():
    num = 7
    mask = []
    for ii in range(5, num + 5):
        mii = np.zeros(200, dtype=bool)
        mii[:ii] = True
        mask.append(mii.reshape(20, 10))
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "mask": mask}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_mask_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "mask" in events.keys()
        # Masks are stored as uint8
        assert np.allclose(events["mask"][6], mask[6]*255)
        assert events["mask"][1].shape == (20, 10)


def test_bulk_logs():
    log = ["This is a test log that contains two lines.",
           "This is the second line.",
           ]
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_logs_")
    hw = RTDCWriter(rtdc_file)
    hw.store_log("testlog", log)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        outlog = rtdc_data["logs"]["testlog"]
        for ii in range(len(outlog)):
            if isinstance(outlog[ii], bytes):
                # h5py 3 reads strings as bytes by default
                outii = outlog[ii].decode("utf-8")
            else:
                outii = outlog[ii]
            assert outii == log[ii]


def test_bulk_trace():
    num = 20
    trace = {"fl1_median": np.arange(num * 111).reshape(num, 111),
             "fl1_raw": 13 + np.arange(num * 111).reshape(num, 111),
             }
    data = {"area_um": np.linspace(100.7, 110.9, num),
            "trace": trace}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_bulk_trace_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "trace" in events.keys()
        assert np.allclose(events["trace"]["fl1_raw"], trace["fl1_raw"])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contour_from_hdf5():
    ds1 = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert ds1["contour"].shape == (5, np.nan, 2)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    with RTDCWriter(f1) as hw:
        hw.store_metadata({"setup": ds1.config["setup"],
                           "experiment": ds1.config["experiment"]})
        hw.store_feature("deform", ds1["deform"])
        hw.store_feature("contour", ds1["contour"])

    ds2 = new_dataset(f1)
    assert ds2["contour"].shape == (5, np.nan, 2)


def test_empty_array():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with RTDCWriter(rtdc_file) as hw:
        with pytest.raises(ValueError, match="Empty data object for 'deform'"):
            hw.store_feature("deform", [])


def test_index_increment():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with RTDCWriter(rtdc_file) as hw:
        hw.store_metadata({"experiment": {"sample": "test",
                                          "run index": 1}})
        hw.store_feature("index", np.arange(1, 11))
        hw.store_feature("index", np.arange(1, 11))

    with new_dataset(rtdc_file) as ds:
        assert np.all(ds["index"] == np.arange(1, 21))


def test_index_increment2():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with RTDCWriter(rtdc_file) as hw:
        hw.store_metadata({"experiment": {"sample": "test",
                                          "run index": 1}})
        # actually, RTDCWriter only uses the shape!
        hw.store_feature("index", np.arange(10, 20))
        hw.store_feature("index", np.arange(50, 60))

    with new_dataset(rtdc_file) as ds:
        assert np.all(ds["index"] == np.arange(1, 21))


def test_index_online_no_increment():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with RTDCWriter(rtdc_file) as hw:
        hw.store_metadata({"experiment": {"sample": "test",
                                          "run index": 1}})
        hw.store_feature("index_online", np.arange(10))
        hw.store_feature("index_online", np.arange(10))

    with new_dataset(rtdc_file) as ds:
        assert np.all(ds["index_online"] == np.concatenate([np.arange(10),
                                                            np.arange(10)]))


def test_data_error():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with pytest.raises(ValueError, match="unknown"):
        RTDCWriter(rtdc_file, mode="unknown")

    hw = RTDCWriter(rtdc_file)
    with pytest.raises(ValueError, match="area_undefined"):
        hw.store_feature("area_undefined", np.linspace(100.7, 110.9, 100))

    with pytest.raises(ValueError, match="fl_unknown"):
        hw.store_feature("trace", {"fl_unknown": np.arange(10)})


def test_logs_append():
    log1 = ["This is a test log that contains two lines.",
            "This is the second line.",
            ]
    log2 = ["These are other logging events.",
            "They are appended to the log.",
            "And may have different lengths."
            ]
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_append_logs_")
    hw = RTDCWriter(rtdc_file, mode="reset")
    hw.store_log("testlog", log1)
    hw.store_log("testlog", log2)

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        outlog = rtdc_data["logs"]["testlog"]
        for ii in range(len(outlog)):
            if isinstance(outlog[ii], bytes):
                # h5py 3 reads strings as bytes by default
                outii = outlog[ii].decode("utf-8")
            else:
                outii = outlog[ii]
            assert outii == (log1 + log2)[ii]


def test_meta():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    meta = {"setup": {
        "channel width": 20,
        "chip region": "Channel",  # should be made lower-case
    },
        "online_contour": {
        "no absdiff": "True",  # should be converted to bool
        "image blur": 3.0,
    },
    }
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_meta_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
        hw.store_metadata(meta)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        abool = rtdc_data.attrs["online_contour:no absdiff"]
        assert abool
        assert isinstance(abool, (bool, np.bool_))
        anint = rtdc_data.attrs["online_contour:image blur"]
        assert isinstance(anint, numbers.Integral)
        assert rtdc_data.attrs["setup:channel width"] == 20
        assert rtdc_data.attrs["setup:chip region"] == "channel"


def test_meta_bytes():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    meta = {
        "setup": {
            "channel width": 20,
            "chip region": b"channel"  # bytes should be converted to str
        },
        "experiment": {
            "date": b"2020-08-12"  # bytes should be converted to str
        }}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_meta_")
    with RTDCWriter(rtdc_file) as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])
        hw.store_metadata(meta)
    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        assert rtdc_data.attrs["setup:channel width"] == 20
        assert rtdc_data.attrs["setup:chip region"] == "channel"
        assert rtdc_data.attrs["experiment:date"] == "2020-08-12"


def test_meta_error():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_meta_")

    meta1 = {"rediculous_section": {"a": 4}}
    hw = RTDCWriter(rtdc_file)
    with pytest.raises(ValueError, match="rediculous_section"):
        hw.store_metadata(meta1)

    meta2 = {"setup": {"rediculous_key": 4}}
    with pytest.raises(ValueError, match="rediculous_key"):
        hw.store_metadata(meta2)


def test_meta_no_analysis():
    """The "filtering" section should not be written to the dataset"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_meta_no_analysis")

    meta1 = {"filtering": {"enable filters": True}}
    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("area_um", np.linspace(100.7, 110.9, 100))
        with pytest.raises(ValueError, match="filtering"):
            hw.store_metadata(meta1)


def test_mode():
    data = {"area_um": np.linspace(100.7, 110.9, 100)}
    data2 = {"deform": np.linspace(.7, .8, 100)}

    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_")
    with RTDCWriter(rtdc_file, mode="reset") as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])

    with RTDCWriter(rtdc_file, mode="append") as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data1:
        events1 = rtdc_data1["events"]
        assert "area_um" in events1.keys()
        assert len(events1["area_um"]) == 2 * len(data["area_um"])

    with RTDCWriter(rtdc_file, mode="replace") as hw:
        for feat in data:
            hw.store_feature(feat, data[feat])

    with RTDCWriter(rtdc_file, mode="replace") as hw:
        for feat in data2:
            hw.store_feature(feat, data2[feat])

    with h5py.File(rtdc_file, mode="r") as rtdc_data2:
        events2 = rtdc_data2["events"]
        assert "area_um" in events2.keys()
        assert "deform" in events2.keys()
        assert len(events2["area_um"]) == len(data["area_um"])


def test_nan_array():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with RTDCWriter(rtdc_file) as hw:
        with pytest.warns(RuntimeWarning) as record:
            hw.store_feature("deform", np.full(100, np.nan))

    assert len(record) == 3
    assert record[0].message.args[0] == "All-NaN axis encountered"
    assert record[1].message.args[0] == "All-NaN axis encountered"
    assert record[2].message.args[0] == "Mean of empty slice"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_non_scalar_bad_shape():
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    exppath = h5path.with_name("exported.rtdc")

    register_temporary_feature("peterpan", is_scalar=False)
    with RTDCWriter(exppath, mode="append") as hw:
        data = np.arange(10 * 3 * 5).reshape(10, 3, 5)
        # This should work
        hw.store_feature("peterpan", data, shape=(3, 5))
        # This should not work
        with pytest.raises(ValueError, match="Bad shape"):
            hw.store_feature("peterpan", data, shape=(3, 6))
    deregister_temporary_feature("peterpan")  # cleanup


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_open_from_h5group():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with h5py.File(path, "a") as h5:
        with dclab.RTDCWriter(h5) as hw:
            assert not hw.owns_path
            hw.store_log("peter", "hans")
        # do it again (to make sure h5 is not closed)
        with dclab.RTDCWriter(h5) as hw:
            assert not hw.owns_path
            hw.store_log("peter", "gert")
    # sanity check
    with dclab.new_dataset(path) as ds:
        assert "".join(ds.logs["peter"]) == "hansgert"


def test_real_time():
    # Create huge array
    n = 116
    # Writing 10 images at a time is faster than writing one image at a time
    m = 4
    assert n // m == np.round(n / m)
    shx = 48
    shy = 32
    contours = [np.arange(20).reshape(10, 2)] * m
    images = np.zeros((m, shy, shx), dtype=np.uint8)
    masks = np.zeros((m, shy, shx), dtype=np.bool_)
    traces = {"fl1_median": np.arange(m * 55).reshape(m, 55)}
    axis1 = np.linspace(0, 1, m)
    axis2 = np.arange(float(m))
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_realtime_")

    with RTDCWriter(rtdc_file, mode="reset") as hw:
        # simulate real time and write one image at a time
        for ii in range(n // m):
            # print(ii)
            num_img = np.copy(images) + ii

            data = {"area_um": axis1,
                    "area_cvx": axis2,
                    "image": num_img,
                    "contour": contours,
                    "mask": masks,
                    "trace": traces}
            for feat in data:
                hw.store_feature(feat, data[feat])

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert events["image"].shape == (n, shy, shx)
        assert events["area_um"].shape == (n,)
        assert events["contour"]["0"].shape == (10, 2)
        assert events["trace"]["fl1_median"].shape == (n, 55)
        assert np.dtype(events["area_um"]) == float
        assert np.dtype(events["area_cvx"]) == float


def test_real_time_single():
    # Create huge array
    n = 33
    shx = 30
    shy = 10
    image = np.zeros((shy, shx), dtype=np.uint8)
    mask = np.zeros((shy, shx), dtype=np.bool_)
    contour = np.arange(22).reshape(11, 2)
    trace = {"fl1_median": np.arange(43)}

    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_realtime_single_")
    with RTDCWriter(rtdc_file, mode="reset") as hw:
        # simulate real time and write one event at a time
        for ii in range(n):
            data = {"area_um": ii * .1,
                    "area_cvx": ii * 5.,
                    "image": image * ii,
                    "contour": contour,
                    "mask": mask,
                    "trace": trace}
            for feat in data:
                hw.store_feature(feat, data[feat])
            hw.store_log("log1", f"line {ii}")

    # Read the file:
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert events["image"].shape == (n, shy, shx)
        assert events["area_um"].shape == (n,)
        assert events["contour"]["0"].shape == (11, 2)
        assert events["trace"]["fl1_median"].shape == (n, 43)
        assert np.dtype(events["area_um"]) == float
        assert np.dtype(events["area_cvx"]) == float
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == n


def test_rectify_metadata_ignore_empty_image():
    # test introduced in 0.39.7
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with h5py.File(rtdc_file, "a") as h5:
        h5.require_group("events")
        h5["events"].require_dataset(name="image",
                                     shape=(0, 80, 100),
                                     dtype=np.uint8)
        h5["events/deform"] = np.linspace(.1, .12, 7)

        # Initialize writer
        hw = RTDCWriter(h5, mode="append")
        # in previous versions, this did not work, because of the empty trace
        hw.rectify_metadata()
        # make sure that something happened
        assert h5.attrs["experiment:event count"] == 7


def test_rectify_metadata_ignore_empty_trace():
    # test introduced in 0.39.7
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_error_")
    with h5py.File(rtdc_file, "a") as h5:
        h5.require_group("events")
        h5.require_group("events/trace")
        h5["events/deform"] = np.linspace(.1, .12, 7)

        # Initialize writer
        hw = RTDCWriter(h5, mode="append")
        # in previous versions, this did not work, because of the empty trace
        hw.rectify_metadata()
        # make sure that something happened
        assert h5.attrs["experiment:event count"] == 7


def test_replace_contour():
    num = 7
    contour = []
    contour2 = []
    for ii in range(5, num + 5):
        cii = np.arange(2 * ii).reshape(2, ii)
        contour.append(cii)
        contour2.append(cii * 2)

    data1 = {"area_um": np.linspace(100.7, 110.9, num),
             "contour": contour}
    data2 = {"contour": contour2}
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_contour_")

    with RTDCWriter(rtdc_file) as hw:
        for feat in data1:
            hw.store_feature(feat, data1[feat])

    with RTDCWriter(rtdc_file, mode="replace") as hw:
        for feat in data2:
            hw.store_feature(feat, data2[feat])

    # verify
    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        events = rtdc_data["events"]
        assert "contour" in events.keys()
        assert not np.allclose(events["contour"]["6"], contour[6])
        assert np.allclose(events["contour"]["6"], contour2[6])


def test_replace_logs():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_logs_")

    with RTDCWriter(rtdc_file) as hw:
        hw.store_log("log1", ["hans", "und"])

    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == 2

    with RTDCWriter(rtdc_file, mode="replace") as hw:
        hw.store_log("log1", ["peter"])

    with h5py.File(rtdc_file, mode="r") as rtdc_data:
        logs = rtdc_data["logs"]
        assert len(logs["log1"]) == 1


def test_scalar_ufuncs_attrs():
    """Since version 0.46.7 we support caching min/max/mean as h5 attributes"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_logs_")
    deform = np.linspace(.1, .2, 254)
    area_um = np.linspace(200, 500, 254)
    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", deform)
        hw.store_feature("area_um", area_um)

    with h5py.File(rtdc_file) as h5:
        assert np.allclose(h5["events/deform"].attrs["min"], 0.1)
        assert np.allclose(h5["events/deform"].attrs["max"], 0.2)
        assert np.allclose(h5["events/deform"].attrs["mean"], 0.15)

        assert np.allclose(h5["events/area_um"].attrs["min"], 200)
        assert np.allclose(h5["events/area_um"].attrs["max"], 500)
        assert np.allclose(h5["events/area_um"].attrs["mean"], 350)


def test_scalar_ufuncs_attrs_append():
    """Since version 0.46.7 we support caching min/max/mean as h5 attributes"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_replace_logs_")
    deform_a = np.linspace(.1, .2, 254)
    area_um_a = np.linspace(200, 500, 254)
    deform_b = np.linspace(.05, .1, 123)
    area_um_b = np.linspace(200, 700, 123)

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", deform_a)
        hw.store_feature("area_um", area_um_a)

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", deform_b)
        hw.store_feature("area_um", area_um_b)

    with h5py.File(rtdc_file) as h5:
        assert np.allclose(h5["events/deform"].attrs["min"], 0.05)
        assert np.allclose(h5["events/deform"].attrs["max"], 0.2)
        assert np.allclose(h5["events/deform"].attrs["mean"],
                           0.12553050397877985)

        assert np.allclose(h5["events/area_um"].attrs["min"], 200)
        assert np.allclose(h5["events/area_um"].attrs["max"], 700)
        assert np.allclose(h5["events/area_um"].attrs["mean"],
                           382.6259946949602)


def test_tables_recarray():
    """Write a numpy recarray table to an .rtdc file"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_tables_recarray")

    # generate a table
    columns = ["bread", "beer", "chocolate"]
    ds_dt = np.dtype({'names': columns,
                      'formats': [np.float64] * len(columns)})
    tab_data = np.zeros((10, len(columns)))
    tab_data[:, 0] = np.arange(10)
    tab_data[:, 1] = 1000
    tab_data[:, 2] = np.linspace(np.pi, 2*np.pi, 10)
    rec_arr = np.rec.array(tab_data, dtype=ds_dt)
    # sanity check
    assert np.all(rec_arr["bread"][:].flatten() == np.arange(10))
    assert np.all(rec_arr["beer"][:].flatten() == 1000)
    assert np.all(rec_arr["chocolate"][:].flatten() == np.linspace(
        np.pi, 2 * np.pi, 10))

    # Write the table to the .rtdc file
    with RTDCWriter(rtdc_file) as hw:
        hw.store_table("food", rec_arr)

    # Check whether this worked
    with h5py.File(rtdc_file) as h5:
        assert np.all(h5["tables"]["food"]["bread"].flat == np.arange(10))


def test_tables_recarray_from_h5array():
    """Write a numpy recarray table to an .rtdc file read from another file"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_tables_recarray")
    rtdc_file2 = tempfile.mktemp(suffix=".rtdc",
                                 prefix="dclab_test_tables_recarray")

    # generate a table
    columns = ["bread", "beer", "chocolate"]
    ds_dt = np.dtype({'names': columns,
                      'formats': [np.float64] * len(columns)})
    tab_data = np.zeros((10, len(columns)))
    tab_data[:, 0] = np.arange(10)
    tab_data[:, 1] = 1000
    tab_data[:, 2] = np.linspace(np.pi, 2*np.pi, 10)
    rec_arr = np.rec.array(tab_data, dtype=ds_dt)
    # sanity check
    assert np.all(rec_arr["bread"][:].flatten() == np.arange(10))
    assert np.all(rec_arr["beer"][:].flatten() == 1000)
    assert np.all(rec_arr["chocolate"][:].flatten() == np.linspace(
        np.pi, 2 * np.pi, 10))

    # Write the table to the .rtdc file
    with RTDCWriter(rtdc_file) as hw:
        hw.store_table("food", rec_arr)

    # Write a copy of that table directly to another HDF5 file
    with h5py.File(rtdc_file) as h5, RTDCWriter(rtdc_file2) as hw2:
        hw2.store_table("food2", h5["tables"]["food"])

    # Check whether that worked
    with h5py.File(rtdc_file2) as h52:
        assert np.all(h52["tables"]["food2"]["bread"].flat == np.arange(10))


def test_tables_recarray_from_dict():
    """Write a numpy recarray table to an .rtdc file"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_tables_recarray")

    # generate a table
    tab_data = {
        "bread": np.arange(10),
        "beer": np.full(10, 1000),
        "chocolate": np.linspace(np.pi, 2*np.pi, 10)
    }

    # Write the table to the .rtdc file
    with RTDCWriter(rtdc_file) as hw:
        hw.store_table("food", tab_data)

    # Check whether this worked
    with h5py.File(rtdc_file) as h5:
        assert np.all(h5["tables"]["food"]["bread"].flat == np.arange(10))


def test_tables_recarray_from_dict_with_lists():
    """Write a numpy recarray table to an .rtdc file"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_tables_recarray")

    # generate a table
    tab_data = {
        "bread": np.arange(10).tolist(),
        "beer": np.full(10, 1000).tolist(),
        "chocolate": np.linspace(np.pi, 2*np.pi, 10).tolist()
    }

    # Write the table to the .rtdc file
    with RTDCWriter(rtdc_file) as hw:
        hw.store_table("food", tab_data)

    # Check whether this worked
    with h5py.File(rtdc_file) as h5:
        assert np.all(h5["tables"]["food"]["bread"].flat == np.arange(10))


def test_tables_recarray_invalid():
    """Write a numpy recarray table to an .rtdc file"""
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_tables_recarray")

    # generate an invalid table
    tab_data = [np.arange(10),
                np.full(10, 1000),
                np.linspace(np.pi, 2*np.pi, 10),
                ]

    # Write the table to the .rtdc file
    with RTDCWriter(rtdc_file) as hw:
        with pytest.raises(NotImplementedError, match="list"):
            hw.store_table("food", tab_data)


def test_version_branding_1_write_single_version():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_branding_")

    this_version = f"dclab {dclab.__version__}"

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", np.linspace(0.01, 0.02, 10))

    with h5py.File(rtdc_file) as h5:
        assert h5.attrs["setup:software version"] == this_version


def test_version_branding_2_dont_override_initial_version():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_branding_")

    this_version = f"dclab {dclab.__version__}"

    with h5py.File(rtdc_file, "a") as h5:
        h5.attrs["setup:software version"] = "Shape-In 1.2.3"

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", np.linspace(0.01, 0.02, 10))

    expected_version = f"Shape-In 1.2.3 | {this_version}"

    with h5py.File(rtdc_file) as h5:
        assert h5.attrs["setup:software version"] == expected_version


def test_version_branding_3_use_old_version_manual():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_branding_")

    this_version = f"dclab {dclab.__version__}"

    with h5py.File(rtdc_file, "a") as h5:
        h5.attrs["setup:software version"] = "Shape-In 1.2.3"

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", np.linspace(0.01, 0.02, 10))
        hw.version_brand(old_version="Peter 1.0")

    expected_version = f"Peter 1.0 | {this_version}"

    with h5py.File(rtdc_file) as h5:
        assert h5.attrs["setup:software version"] == expected_version


def test_version_branding_4_dont_write_attribute():
    rtdc_file = tempfile.mktemp(suffix=".rtdc",
                                prefix="dclab_test_branding_")

    this_version = f"dclab {dclab.__version__}"

    with h5py.File(rtdc_file, "a") as h5:
        h5.attrs["setup:software version"] = "Shape-In 1.2.3"

    with RTDCWriter(rtdc_file) as hw:
        hw.store_feature("deform", np.linspace(0.01, 0.02, 10))
        hw.version_brand(old_version="Peter 1.0", write_attribute=False)

    expected_version = f"Shape-In 1.2.3 | {this_version}"

    with h5py.File(rtdc_file) as h5:
        assert h5.attrs["setup:software version"] == expected_version



import collections
import warnings

import numpy as np

from ...util import copy_if_needed
from .mapper import map_indices_child2parent


class ChildBase(object):
    def __init__(self, child):
        self.child = child

    def __len__(self):
        return len(self.child)


class ChildContour(ChildBase):
    def __init__(self, child):
        super(ChildContour, self).__init__(child)
        self.shape = (len(child), np.nan, 2)
        # Note that since the contour has variable lengths, we cannot
        # implement an __array__ method here.

    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=idx)
        hp = self.child.hparent
        return hp["contour"][pidx]

    @property
    def dtype(self):
        return self.child.hparent["contour"].dtype


class ChildNDArray(ChildBase):
    def __init__(self, child, feat):
        super(ChildNDArray, self).__init__(child)
        self.feat = feat

    def __array__(self, dtype=None, copy=copy_if_needed, *args, **kwargs):
        warnings.warn("Please avoid calling the `__array__` method of the "
                      "`ChildNDArray`. It may consume a lot of memory. "
                      "Consider using a generator instead.",
                      UserWarning)
        return np.asarray(self[:], dtype=dtype, *args, **kwargs)

    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=idx)
        hp = self.child.hparent
        return hp[self.feat][pidx]

    @property
    def dtype(self):
        return self.child.hparent[self.feat].dtype

    @property
    def shape(self):
        hp = self.child.hparent
        return tuple([len(self)] + list(hp[self.feat][0].shape))


class ChildScalar(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(self, child, feat):
        self.child = child
        self.feat = feat
        self._array = None
        # ufunc metadata attribute cache
        self._ufunc_attrs = {}
        self.ndim = 1  # matplotlib might expect this from an array

    def __array__(self, dtype=None, copy=copy_if_needed, *args, **kwargs):
        if self._array is None:
            hparent = self.child.hparent
            filt_arr = hparent.filter.all
            self._array = hparent[self.feat][filt_arr]
        return np.array(self._array, dtype=dtype, copy=copy, *args, **kwargs)

    def __getitem__(self, idx):
        return self.__array__()[idx]

    def __len__(self):
        return len(self.child)

    def _fetch_ufunc_attr(self, uname, ufunc):
        """A wrapper for calling functions on the scalar feature data

        If the ufunc is computed, it is cached permanently in
        self._ufunc_attrs
        """
        val = self._ufunc_attrs.get(uname, None)
        if val is None:
            val = ufunc(self.__array__())
            self._ufunc_attrs[uname] = val
        return val

    def max(self, *args, **kwargs):
        return self._fetch_ufunc_attr("max", np.nanmax)

    def mean(self, *args, **kwargs):
        return self._fetch_ufunc_attr("mean", np.nanmean)

    def min(self, *args, **kwargs):
        return self._fetch_ufunc_attr("min", np.nanmin)

    @property
    def shape(self):
        return len(self),


class ChildTrace(collections.UserDict):
    @property
    def shape(self):
        # set proper shape (#117)
        key0 = sorted(self.keys())[0]
        return tuple([len(self)] + list(self[key0].shape))


class ChildTraceItem(ChildBase):
    def __init__(self, child, flname):
        super(ChildTraceItem, self).__init__(child)
        self.flname = flname

    def __array__(self, dtype=None, copy=copy_if_needed, *args, **kwargs):
        warnings.warn("Please avoid calling the `__array__` method of the "
                      "`ChildTraceItem`. It may consume a lot of memory. "
                      "Consider using a generator instead.",
                      UserWarning)
        return np.asarray(self[:], dtype=dtype, *args, **kwargs)

    def __getitem__(self, idx):
        pidx = map_indices_child2parent(child=self.child,
                                        child_indices=idx)
        hp = self.child.hparent
        return hp["trace"][self.flname][pidx]

    @property
    def dtype(self):
        hp = self.child.hparent
        return hp["trace"][self.flname].dtype

    @property
    def shape(self):
        hp = self.child.hparent
        return len(self), hp["trace"][self.flname].shape[1]



"""Test DCOR format"""
import pathlib

import dclab
from dclab.rtdc_dataset.fmt_dcor import RTDC_DCOR, is_dcor_url
import numpy as np
import pytest

from helper_methods import DCOR_AVAILABLE, retrieve_data


pytest.importorskip("requests")


class MockAPIHandler(dclab.rtdc_dataset.fmt_dcor.api.APIHandler):
    def __init__(self, *args, **kwargs):
        super(MockAPIHandler, self).__init__(*args, **kwargs)
        # We are mocking only API version 2
        self.dcserv_api_version = 2

    def get(self, query, feat=None, trace=None, event=None):
        """Mocks communication with the DCOR API"""
        h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
        with dclab.new_dataset(h5path) as ds:
            if query == "valid":
                return True
            elif query == "metadata":
                return dict(ds.config)
            elif query == "feature_list":
                return []
            elif query == "logs":
                return ds.logs
            elif query == "size":
                return len(ds)
            elif query == "basins":
                return [{
                    "description": "mock basin",
                    "format": "hdf5",
                    "name": "fmt-hdf5_fl_2018",
                    "type": "file",
                    "paths": [str(h5path)],
                }]
            elif query == "tables":
                return ds.tables
            elif query == "trace_list":
                return sorted(ds["trace"].keys())
            else:
                raise ValueError(f"Invalid query '{query}'")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_dcor_base(monkeypatch):
    monkeypatch.setattr(dclab.rtdc_dataset.fmt_dcor.api,
                        "APIHandler",
                        MockAPIHandler)
    with dclab.new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip")) as ds:
        dso = dclab.new_dataset(
            "https://example.com/api/3/action/dcserv?id="
            "b1404eb5-f661-4920-be79-5ff4e85915d1")
        # Trick fmt_dcor into accepting file-based basins
        dso._local_basins_allowed = True

        assert len(dso) == len(ds)
        assert dso.config["setup"]["channel width"] == \
            ds.config["setup"]["channel width"]
        assert ds["area_um"].ndim == 1  # important for matplotlib
        assert np.all(dso["area_um"] == ds["area_um"])
        assert np.all(dso["area_um"] == ds["area_um"])  # test cache
        assert np.all(dso["image"][4] == ds["image"][4])
        assert len(dso["image"]) == len(ds)
        for key in dso._events:
            assert key in ds
        for m, n in zip(dso["mask"], ds["mask"]):
            assert np.all(m == n)
        # compute an ancillary feature
        assert np.all(dso["volume"] == ds["volume"])
        assert np.all(dso["volume"] == ds["volume"])  # test cache
        # trace
        assert sorted(dso["trace"].keys()) == sorted(ds["trace"].keys())
        assert len(dso["trace"]["fl1_raw"]) == len(ds["trace"]["fl1_raw"])
        assert np.all(dso["trace"]["fl1_raw"][1] == ds["trace"]["fl1_raw"][1])
        for t1, t2 in zip(dso["trace"]["fl1_raw"], ds["trace"]["fl1_raw"]):
            assert np.all(t1 == t2)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_data():
    # reticulocytes.rtdc contains contour data
    with dclab.new_dataset("13247dd0-3d8b-711d-a410-468b4de6fb7a") as ds:
        assert np.allclose(ds["circ"][0],
                           0.7309052348136902,
                           rtol=0, atol=1e-5)
        assert np.allclose(ds["area_um"][391], 37.5122, rtol=0, atol=1e-5)
        assert np.all(ds["contour"][24][22] == np.array([87, 61]))
        assert np.median(ds["image"][1]) == 58
        assert np.sum(ds["mask"][11]) == 332
        assert np.sum(ds["mask"][11]) == 332
        assert np.median(ds["trace"]["fl1_raw"][200]) == 183.0
        assert np.sum(ds["trace"]["fl1_median"][2167]) == 183045


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_hash():
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        # hash includes the full URL (path)
        assert ds.hash == "7250277b41b757cbe09647a58e8ca4ce"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_dcor_hierarchy(monkeypatch):
    monkeypatch.setattr(dclab.rtdc_dataset.fmt_dcor.api,
                        "APIHandler",
                        MockAPIHandler)
    dso = dclab.new_dataset("https://example.com/api/3/action/dcserv?id="
                            "b1404eb5-f661-4920-be79-5ff4e85915d5")
    # Trick fmt_dcor into accepting file-based basins
    dso._local_basins_allowed = True

    dsh = dclab.new_dataset(dso)
    assert np.all(dso["area_um"] == dsh["area_um"])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_logs():
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds.logs) >= 2  # there might be others
        assert ds.logs["log"][0] \
               == "[LOG] number of written datasets 0  10:04:05.893"
        assert "dclab-condense" in ds.logs


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_logs_iter():
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds.logs) >= 2  # there might be others
        for key in ds.logs:
            assert key


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_shape_contour():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds["contour"]) == 5000
        assert ds["contour"].shape == (5000, np.nan, 2)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_shape_image():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds["image"]) == 5000
        assert ds["image"].shape == (5000, 80, 250)
        assert ds["image"][0].shape == (80, 250)
        assert ds["image"][0][0].shape == (250,)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_shape_mask():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds["mask"]) == 5000
        assert ds["mask"].shape == (5000, 80, 250)
        assert ds["mask"][0].shape == (80, 250)
        assert ds["mask"][0][0].shape == (250,)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_shape_trace():
    # calibration beads
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert len(ds["trace"]) == 6
        assert ds["trace"].shape == (6, 5000, 177)
        assert len(ds["trace"]["fl1_raw"]) == 5000
        assert ds["trace"]["fl1_raw"].shape == (5000, 177)
        assert len(ds["trace"]["fl1_raw"][0]) == 177
        assert len(ds["trace"]["fl1_raw"][0]) != 5000
        assert ds["trace"]["fl1_raw"][0].shape == (177,)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("idxs", [slice(0, 5, 2),
                                  np.array([0, 2, 4]),
                                  [0, 2, 4]
                                  ])
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_slicing_contour(idxs):
    """Test slicing of contour data"""
    # reticulocytes.rtdc contains contour data
    with dclab.new_dataset("13247dd0-3d8b-711d-a410-468b4de6fb7a") as ds:

        data_ref = [
            ds["contour"][0],
            ds["contour"][2],
            ds["contour"][4],
        ]

        data_sliced = ds["contour"][idxs]

        assert np.all(data_sliced[0] == data_ref[0])
        assert np.all(data_sliced[1] == data_ref[1])
        assert np.all(data_sliced[2] == data_ref[2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("feat", ["image", "mask"])
@pytest.mark.parametrize("idxs", [slice(0, 5, 2),
                                  np.array([0, 2, 4]),
                                  [0, 2, 4]
                                  ])
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_slicing_image_mask(feat, idxs):
    """Test slicing of image/mask data"""
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        data_ref = [
            ds[feat][0],
            ds[feat][2],
            ds[feat][4],
        ]

        data_sliced = ds[feat][idxs]

        assert np.all(data_sliced[0] == data_ref[0])
        assert np.all(data_sliced[1] == data_ref[1])
        assert np.all(data_sliced[2] == data_ref[2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("idxs", [slice(0, 5, 2),
                                  np.array([0, 2, 4]),
                                  [0, 2, 4]
                                  ])
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_dcor_slicing_trace(idxs):
    """Test slicing of trace data"""
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        data_ref = [
            ds["trace"]["fl1_raw"][0],
            ds["trace"]["fl1_raw"][2],
            ds["trace"]["fl1_raw"][4],
        ]

        data_sliced = ds["trace"]["fl1_raw"][idxs]

        assert np.all(data_sliced[0] == data_ref[0])
        assert np.all(data_sliced[1] == data_ref[1])
        assert np.all(data_sliced[2] == data_ref[2])


@pytest.mark.parametrize("target,kwargs", [
    # HTTPS
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": True,
      "host": "example.com"}
     ),
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": True,
      "host": "example.com"}
     ),
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": None,
      "host": "example.com"}
     ),
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": None,
      "host": "example.com"}
     ),
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": None,
      # sneak in a scheme into host
      "host": "https://example.com"}
     ),
    ("https://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "https://example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": None,
      # `host` does not override things
      "host": "example2.com"}
     ),
    # HTTP
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": False,
      "host": "example.com"}
     ),
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "https://example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": False,
      "host": "example.com"}
     ),
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": False,
      "host": "example.com"}
     ),
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "http://example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": None,
      "host": "example.com"}
     ),
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": False,
      # sneak in a scheme into host
      "host": "https://example.com"}
     ),
    ("http://example.com/api/3/action/dcserv?"
     "id=b1404eb5-f661-4920-be79-5ff4e85915d5",
     {"url": "http://example.com/api/3/action/dcserv?id="
             "b1404eb5-f661-4920-be79-5ff4e85915d5",
      "use_ssl": False,
      # `host` does not override things
      "host": "example2.com"}
     ),
])
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_get_full_url(target, kwargs):
    assert target == RTDC_DCOR.get_full_url(**kwargs)


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_is_dcor_url():
    assert is_dcor_url("2cea205f-2d9d-26d0-b44c-0a11d5379152")
    assert not is_dcor_url("2cea205f-2d9d")
    assert is_dcor_url("https://example.com/api/3/action/dcserv?id="
                       "2cea205f-2d9d-26d0-b44c-0a11d5379152")
    assert is_dcor_url("http://example.com/api/3/action/dcserv?id="
                       "2cea205f-2d9d-26d0-b44c-0a11d5379152")
    assert is_dcor_url("example.com/api/3/action/dcserv?id="
                       "2cea205f-2d9d-26d0-b44c-0a11d5379152")
    assert not is_dcor_url(
        pathlib.Path("example.com/api/3/action/dcserv?id="
                     "2cea205f-2d9d-26d0-b44c-0a11d5379152"))
    assert not is_dcor_url(2.0)
    assert not is_dcor_url("/home/peter/pan")
    assert not is_dcor_url("https://example.com/api/3/action/dcserv?id="
                           "2cea205f-2d9d-26d0-b44c-")


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="no connection to DCOR")
def test_load_nonexistent_file_issue81():
    """https://github.com/DC-analysis/dclab/issues/81"""
    try:
        dclab.new_dataset("path/does/not/exist.rtdc")
    except FileNotFoundError:
        pass
    else:
        assert False, "Non-existent files should raise FileNotFoundError"



"""DCOR client interface"""
import pathlib
import re

from ...util import hashobj

from ..config import Configuration
from ..core import RTDCBase

from . import api
from .logs import DCORLogs
from .tables import DCORTables


#: Append directories here where dclab should look for certificate bundles
#: for a specific host. The directory should contain files named after the
#: hostname, e.g. "dcor.mpl.mpg.de.cert".
DCOR_CERTS_SEARCH_PATHS = []

#: Regular expression for matching a DCOR resource URL
REGEXP_DCOR_URL = re.compile(
    r"^(https?:\/\/)?"  # scheme
    r"([a-z0-9-\.]*\/?api\/3\/action\/dcserv\?id=)?"  # host with API
    r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$")  # id


class RTDC_DCOR(RTDCBase):
    def __init__(self, url, host="dcor.mpl.mpg.de", api_key="",
                 use_ssl=None, cert_path=None, dcserv_api_version=2,
                 *args, **kwargs):
        """Wrap around the DCOR API

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - `<https://dcor.mpl.mpg.de/api/3/action/dcserv?id=
              b1404eb5-f661-4920-be79-5ff4e85915d5>`_
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=b1404eb5-f
              661-4920-be79-5ff4e85915d5
            - b1404eb5-f661-4920-be79-5ff4e85915d5
        host: str
            The default host machine used if the host is not given in `url`
        api_key: str
            API key to access private resources
        use_ssl: bool
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        cert_path: pathlib.Path
            The (optional) path to a server CA bundle; this should only
            be necessary for DCOR instances in the intranet with a custom
            CA or for certificate pinning.
        dcserv_api_version: int
            Version of the dcserv API to use. In version 0.13.2 of
            ckanext-dc_serve, version 2 was introduced which entails
            serving an S3-basin-only dataset.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            Full URL to the DCOR resource
        """
        if not api.REQUESTS_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `requests` required for DCOR format!")

        super(RTDC_DCOR, self).__init__(*args, **kwargs)

        self._hash = None
        self.path = RTDC_DCOR.get_full_url(url, use_ssl, host)

        if cert_path is None:
            cert_path = get_server_cert_path(get_host_from_url(self.path))

        self.api = api.APIHandler(url=self.path,
                                  api_key=api_key,
                                  cert_path=cert_path,
                                  dcserv_api_version=dcserv_api_version)

        # Parse configuration
        self.config = Configuration(cfg=self.api.get(query="metadata"))

        # Lazy logs
        self.logs = DCORLogs(self.api)

        # Lazy tables
        self.tables = DCORTables(self.api)

        # Get size
        size = self.config["experiment"].get("event count")
        if size is None:
            size = int(self.api.get(query="size"))
        self._size = size

        self.title = f"{self.config['experiment']['sample']} - " \
            + f"M{self.config['experiment']['run index']}"

    def __len__(self):
        return self._size

    @property
    def hash(self):
        """Hash value based on file name and content"""
        if self._hash is None:
            tohash = [self.path]
            self._hash = hashobj(tohash)
        return self._hash

    @staticmethod
    def get_full_url(url, use_ssl, host=None):
        """Return the full URL to a DCOR resource

        Parameters
        ----------
        url: str
            Full URL or resource identifier; valid values are

            - https://dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-
              df12-4299-aa2e-089e390aafd5'
            - dcor.mpl.mpg.de/api/3/action/dcserv?id=caab96f6-df12-
              4299-aa2e-089e390aafd5
            - caab96f6-df12-4299-aa2e-089e390aafd5
        use_ssl: bool or None
            Set this to False to disable SSL (should only be used for
            testing). Defaults to None (does not force SSL if the URL
            starts with "http://").
        host: str
            Use this host if it is not specified in `url`
        """
        if use_ssl is None:
            if url.startswith("http://"):
                # user wanted it that way
                scheme = "http"
            else:
                scheme = "https"
        elif use_ssl:
            scheme = "https"
        else:
            scheme = "http"
        if url.count("://"):
            base = url.split("://", 1)[1]
        else:
            base = url
        # determine the api_path and the netloc
        if base.count("/"):
            netloc, api_path = base.split("/", 1)
        else:
            netloc = None  # default to `host`
            api_path = "api/3/action/dcserv?id=" + base
        # remove https from host string (user convenience)
        if host is not None:
            host = host.split("://")[-1]

        netloc = host if netloc is None else netloc
        new_url = f"{scheme}://{netloc}/{api_path}"
        return new_url

    def basins_get_dicts(self):
        """Return list of dicts for all basins defined in `self.h5file`"""
        try:
            basins = self.api.get(query="basins")
        except api.DCORAccessError:
            # TODO: Do not catch this exception when all DCOR instances
            #       implement the 'basins' query.
            # This means that the server does not implement the 'basins' query.
            basins = []
        return basins


def get_host_from_url(url):
    """Extract the hostname from a URL"""
    return url.split("://")[1].split("/")[0]


def get_server_cert_path(host):
    """Return server certificate bundle for DCOR `host`"""

    for path in DCOR_CERTS_SEARCH_PATHS:
        path = pathlib.Path(path)
        cert_path = path / f"{host}.cert"
        if cert_path.exists():
            break
    else:
        # use default certificate bundle
        cert_path = api.requests.certs.where()

    return cert_path


def is_dcor_url(string):
    if not isinstance(string, str):
        return False
    else:
        return REGEXP_DCOR_URL.match(string.strip())



import json
from os.path import join
import pathlib
import tempfile

import h5py
import numpy as np
import pytest

import dclab
from dclab import dfn, new_dataset, RTDCWriter
from dclab.rtdc_dataset.export import store_filtered_feature

from helper_methods import example_data_dict, retrieve_data


def test_hdf5():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"])
    assert np.allclose(ds2["deform"], ds1["deform"])
    assert np.allclose(ds2["time"], ds1["time"])
    assert np.allclose(ds2["frame"], ds1["frame"])
    assert np.allclose(ds2["fl3_width"], ds1["fl3_width"])


def test_hdf5_duplicate_feature_for_export():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    keys.append("area_um")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"])
    assert np.allclose(ds2["deform"], ds1["deform"])
    assert np.allclose(ds2["time"], ds1["time"])
    assert np.allclose(ds2["frame"], ds1["frame"])
    assert np.allclose(ds2["fl3_width"], ds1["fl3_width"])


def test_hdf5_contour_image_trace():
    n = 65
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False)
    ds2 = dclab.new_dataset(f1)

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["contour"][ii] == ds2["contour"][ii])

    for key in dfn.FLUOR_TRACES:
        assert np.all(ds1["trace"][key] == ds2["trace"][key])


def test_hdf5_contour_image_trace_large():
    """Same test for large event numbers (to test chunking)"""
    n = 653
    keys = ["contour", "image", "trace"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False)
    ds2 = dclab.new_dataset(f1)

    # the trace may have negative values, it's int16, not uint16.
    assert ds1["trace"]["fl1_median"][0][0] == -1, "sanity check"

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["contour"][ii] == ds2["contour"][ii])

    for key in dfn.FLUOR_TRACES:
        assert np.all(ds1["trace"][key] == ds2["trace"][key])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_contour_from_hdf5():
    ds1 = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert ds1["contour"].shape == (5, np.nan, 2)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image.rtdc")
    ds1.export.hdf5(f1, ["contour"], filtered=False)
    ds2 = dclab.new_dataset(f1)
    assert ds2["contour"].shape == (5, np.nan, 2)


def test_hdf5_filtered():
    n = 10
    keys = ["area_um", "image"]
    ddict = example_data_dict(size=n, keys=keys)
    ddict["image"][3] = np.arange(10 * 20, dtype=np.uint8).reshape(10, 20) + 22

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()
    fta = ds1.filter.manual.copy()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert ds1 != ds2
    assert np.allclose(ds2["area_um"], ds1["area_um"][fta])
    assert np.allclose(ds2["image"][2], ds1["image"][3])
    assert np.all(ds2["image"][2] != ds1["image"][2])


def test_hdf5_filtered_empty():
    n = 10
    keys = ["area_um", "image"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    # Remove all events
    ds1.filter.manual[:] = False
    ds1.apply_filter()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    with pytest.warns(UserWarning, match="No data to export to"):
        ds1.export.hdf5(f1, keys)

    with h5py.File(f1) as h5:
        assert len(h5["events"].keys()) == 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("dataset", ["fmt-hdf5_mask-contour_2018.zip",
                                     "fmt-hdf5_fl-no-contour_2019.zip"])
def test_hdf5_filtered_full(dataset):
    """Compare full export to filter with all-True"""
    tdir = pathlib.Path(tempfile.mkdtemp())
    path_in = tdir / "in.rtdc"

    # first, create a large dataset
    with new_dataset(retrieve_data(dataset)) as ds0:
        num = 100 % len(ds0) + len(ds0)
        ds0.export.hdf5(path_in, features=ds0.features_innate)
        with RTDCWriter(path_in) as hw:
            for _ in range(num):
                for feat in ds0.features_innate:
                    hw.store_feature(feat, ds0[feat])

    # 1. via export unfiltered
    path_out1 = tdir / "out1.rtdc"
    with dclab.new_dataset(path_in) as ds:
        assert len(ds) >= 100
        ds.export.hdf5(path_out1, features=ds.features_innate, filtered=False)

    # 2. via export filtered
    path_out2 = tdir / "out2.rtdc"
    with dclab.new_dataset(path_in) as ds:
        ds.export.hdf5(path_out2, features=ds.features_innate, filtered=True)

    # 3. via RTDC writer
    path_out3 = tdir / "out3.rtdc"
    with dclab.new_dataset(path_in) as ds:
        with RTDCWriter(path_out3) as hw:
            config = ds.config.as_dict(pop_filtering=True)
            hw.store_metadata(config)
            for feat in ds.features_innate:
                hw.store_feature(feat, ds[feat])

    # 4. via export filtered function
    path_out4 = tdir / "out4.rtdc"
    with dclab.new_dataset(path_in) as ds:
        with RTDCWriter(path_out4) as hw:
            config = ds.config.as_dict(pop_filtering=True)
            hw.store_metadata(config)
            for feat in ds.features_innate:
                store_filtered_feature(rtdc_writer=hw,
                                       feat=feat,
                                       data=ds[feat],
                                       filtarr=np.ones(len(ds), dtype=bool))

    with dclab.new_dataset(path_in) as ds0:
        for path in [path_out1, path_out2, path_out3, path_out4]:
            with dclab.new_dataset(path) as dsi:
                for feat in dsi.features_innate:
                    if dfn.scalar_feature_exists(feat):
                        assert np.all(ds0[feat] == dsi[feat])
                    elif feat == "trace":
                        for key in ds0[feat]:
                            assert np.all(np.array(ds0[feat][key])
                                          == np.array(dsi[feat][key]))
                    else:
                        for ii in range(len(ds0)):
                            assert np.all(ds0[feat][ii] == dsi[feat][ii])


def test_hdf5_filtered_index():
    """Make sure that exported index is always re-enumerated"""
    n = 10
    keys = ["area_um", "deform", "index"]
    ddict = example_data_dict(size=n, keys=keys)
    ddict["index"] = np.arange(1, n+1)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.filter.manual[2] = False
    ds1.apply_filter()

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_filtered.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)

    assert len(ds2) == n - 1
    assert np.all(ds2["index"] == np.arange(1, n))
    assert ds2.config["experiment"]["event count"] == n - 1


def test_hdf5_hierarchy_basin_only_export():
    """
    When the dataset that is exported is a hierarchy child,
    then there should be a mapped basin in the output file
    referring to the input data.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.export.hdf5(path_exp,
                            features=[],
                            basins=True,
                            )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for filtering
        assert np.allclose(dse["deform"][2], 0.14734814,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][5:])


def test_hdf5_hierarchy_basin_only_export_with_filters():
    """
    Same as above, only this time we additionally apply filters
    to the child dataset before exporting.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.filter.manual[::2] = False
            dsc.apply_filter()
            dsc.export.hdf5(path_exp,
                            filtered=True,
                            features=[],
                            basins=True,
                            )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2]), "sanity"


def test_hdf5_hierarchy_basin_only_export_with_filters_inception():
    """
    Same as above, here with an export from a hierarchy of depth 2.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()

        with dclab.new_dataset(ds) as dsc:
            dsc.filter.manual[::2] = False
            dsc.apply_filter()
            with dclab.new_dataset(dsc) as dsci:
                dsci.filter.manual[-1] = False
                dsci.apply_filter()
                dsci.export.hdf5(path_exp,
                                 filtered=True,
                                 features=[],
                                 basins=True,
                                 )

    with h5py.File(path_exp) as h5:
        assert "basinmap0" in h5["events"]
        assert "basins" in h5
        assert len(h5["basins"]) == 1

    with dclab.new_dataset(path_exp) as dse, dclab.new_dataset(h5path) as ds:
        assert "image" in dse.features_basin
        assert "mask" in dse.features_basin
        assert "area_um" in dse.features_basin

        assert "image" not in dse.features_innate
        assert "mask" not in dse.features_innate
        assert "area_um" not in dse.features_innate

        # check for inception filtering
        assert np.allclose(dse["deform"][0], 0.15025392,
                           atol=1e-7, rtol=0)
        assert np.all(dse["deform"][:] == ds["deform"][6::2][:-1])
        assert np.all(dse["deform"][:] == ds["deform"][5:][1::2][:-1]), "sane"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_identifier_different():
    """Make sure the identifier is changed upon export"""
    pin = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    pout = pin.with_name("exported.rtdc")
    with new_dataset(pin) as din:
        mid_in = din.get_measurement_identifier()
        assert len(din) == 5
        assert mid_in
        din.config["filtering"]["deform min"] = np.mean(din["deform"])
        din.config["filtering"]["deform max"] = np.max(din["deform"])
        din.apply_filter()
        din.export.hdf5(pout, ["area_um", "deform"], filtered=True)

    with new_dataset(pout) as dout:
        mid_out = dout.get_measurement_identifier()
        assert mid_out.startswith(mid_in)
        assert mid_out != mid_in
        assert len(dout) == 3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_identifier_same_unfiltered():
    """Make sure the identifier is changed upon export"""
    pin = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    pout = pin.with_name("exported.rtdc")
    with new_dataset(pin) as din:
        mid_in = din.get_measurement_identifier()
        assert len(din) == 5
        assert mid_in
        din.config["filtering"]["deform min"] = np.mean(din["deform"])
        din.config["filtering"]["deform max"] = np.max(din["deform"])
        din.apply_filter()
        din.export.hdf5(pout, ["area_um", "deform"], filtered=False)

    with new_dataset(pout) as dout:
        mid_out = dout.get_measurement_identifier()
        assert mid_out == mid_in
        assert len(dout) == 5


def test_hdf5_image_bg():
    n = 65
    keys = ["image", "image_bg"]
    ddict = example_data_dict(size=n, keys=keys)

    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5_image_bg.rtdc")
    ds1.export.hdf5(f1, keys, filtered=False)
    ds2 = dclab.new_dataset(f1)

    for ii in range(n):
        assert np.all(ds1["image"][ii] == ds2["image"][ii])
        assert np.all(ds1["image_bg"][ii] == ds2["image_bg"][ii])


def test_hdf5_index_continuous():
    keys = ["area_um", "deform", "time", "frame", "index"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with RTDCWriter(f1) as hw:
        for feat in keys:
            hw.store_feature(feat, ds1[feat])

    with h5py.File(f1, "r") as h5:
        assert "index" in h5["events"]
        assert np.allclose(h5["events/index"][:], np.arange(1, 21))


def test_hdf5_index_online_continuous():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with RTDCWriter(f1) as hw:
        for feat in keys:
            hw.store_feature(feat, ds1[feat])

    with h5py.File(f1, "r") as h5:
        assert "index_online" in h5["events"]


def test_hdf5_index_online_replaces_index():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)
    with h5py.File(f1, "r") as h5:
        assert "index" not in h5["events"]
        assert "index_online" in h5["events"]


@pytest.mark.parametrize("logs", [True, False])
def test_hdf5_logs(logs):
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000
    ds1.logs["iratrax"] = ["don't", "eat", "brown", "apples"]

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, logs=logs)
    with h5py.File(f1, "r") as h5:
        if logs:
            assert "src_iratrax" in h5.get("logs", {})
        else:
            assert "src_iratrax" not in h5.get("logs", {})


def test_hdf5_logs_meta():
    """Test export of export log #251"""
    keys = ["area_um", "deform", "fl3_width", "frame", "time"]
    ddict = example_data_dict(size=127, keys=keys)
    ds1 = dclab.new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "dclab_test_export_hdf5.rtdc")
    ds1.export.hdf5(f1, keys)

    ds2 = dclab.new_dataset(f1)
    for name in ds2.logs:
        if name.startswith("dclab-export_"):
            kwargs = json.loads("\n".join(ds2.logs[name]))["kwargs"]
            assert kwargs["features"] == keys
            break
    else:
        assert False


def test_hdf5_ml_score():
    data = {"ml_score_ds9": [.80, .31, .12, .01, .59, .40, .52],
            "ml_score_dsc": [.12, .52, .21, .24, .30, .22, .79],
            "ml_score_ent": [.42, .11, .78, .11, .54, .24, .15],
            "ml_score_pic": [.30, .30, .10, .99, .59, .55, .19],
            "ml_score_tng": [.13, .33, .13, .01, .79, .11, .22],
            "ml_score_tos": [.14, .34, .12, .01, .59, .56, .17],
            "ml_score_voy": [.25, .12, .42, .33, .21, .55, .82],
            }
    ds0 = dclab.new_dataset(data)
    ds0.config["experiment"]["sample"] = "test"
    ds0.config["experiment"]["run index"] = 1

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds0.export.hdf5(f1, list(data.keys()), override=True)

    with dclab.new_dataset(f1) as ds:
        assert np.allclose(ds["ml_class"], np.arange(7))
        assert "ml_class" not in ds._events
        assert "ml_class" in ds._ancillaries


def test_hdf5_override():
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)
    ds.config["imaging"]["frame rate"] = 2000

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, keys, override=True)
    try:
        ds.export.hdf5(f1[:-5], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .rtdc and not override!")


@pytest.mark.parametrize("tables", [True, False])
def test_hdf5_tables(tables):
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    # generate a table
    columns = ["bread", "beer", "chocolate"]
    ds_dt = np.dtype({'names': columns,
                      'formats': [np.float64] * len(columns)})
    tab_data = np.zeros((10, len(columns)))
    tab_data[:, 0] = np.arange(10)
    tab_data[:, 1] = 1000
    tab_data[:, 2] = np.linspace(np.pi, 2*np.pi, 10)
    rec_arr = np.rec.array(tab_data, dtype=ds_dt)

    ds1.tables["iratrax"] = rec_arr

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, tables=tables)
    with h5py.File(f1, "r") as h5:
        if tables:
            assert "src_iratrax" in h5.get("tables", {})
        else:
            assert "src_iratrax" not in h5.get("tables", {})

    if tables:
        # reading the file with dclab
        with new_dataset(f1) as ds:
            assert "src_iratrax" in ds.tables
            assert ds.tables["src_iratrax"].has_graphs()
            assert np.all(ds.tables["src_iratrax"] == rec_arr)


def test_hdf5_tables_array_only():
    keys = ["area_um", "deform", "time", "frame", "index_online"]
    ddict = example_data_dict(size=10, keys=keys)
    ds1 = new_dataset(ddict)
    ds1.config["experiment"]["sample"] = "test"
    ds1.config["experiment"]["run index"] = 1
    ds1.config["imaging"]["frame rate"] = 2000

    # generate a table that consists of an array, not a dict-like object
    tab_data = np.random.random((1000, 300))
    ds1.tables["iratrax"] = tab_data

    edest = pathlib.Path(tempfile.mkdtemp())
    f1 = edest / "dclab_test_export_hdf5_1.rtdc"
    ds1.export.hdf5(f1, keys, tables=True)
    with h5py.File(f1, "r") as h5:
        assert "src_iratrax" in h5.get("tables", {})
        assert np.all(h5["tables"]["src_iratrax"] == tab_data)
        assert h5["tables"]["src_iratrax"].attrs["CLASS"] == np.bytes_("IMAGE")

    # reading the file with dclab
    with new_dataset(f1) as ds:
        assert "src_iratrax" in ds.tables
        assert not ds.tables["src_iratrax"].has_graphs()
        assert np.all(ds.tables["src_iratrax"] == tab_data)
        assert ds.tables["src_iratrax"].meta["CLASS"] == np.bytes_("IMAGE")


def test_hdf5_table_from_file(tmp_path):
    ds = new_dataset(retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip"))

    f1 = join(tmp_path, "test.rtdc")
    ds.export.hdf5(f1, ["temp"], tables=True)

    ds2 = new_dataset(f1)
    tab = ds2.tables["src_so2exp_cytoshot_monitor"]

    assert np.allclose(
        tab["focus"][0],
        6.241900073522602,
        atol=1e-7, rtol=0)

    assert tab.meta["COLOR_focus"] == "#a03000"


def test_hdf5_table_from_file_with_unnamed_table(tmp_path):
    path = retrieve_data("fmt-hdf5_raw-cytoshot-exported.zip")
    with RTDCWriter(path) as hw:
        hw.store_table("just_an_array", np.ones((30, 50)),
                       h5_attrs={"heinz": "kunz"})

    # Make sure that worked with tables
    with new_dataset(path) as ds:
        tab = ds.tables["just_an_array"]
        assert not tab.has_graphs()
        assert np.array(tab).shape == (30, 50)
        assert tab.meta["heinz"] == "kunz"

    with new_dataset(path) as ds:
        f1 = join(tmp_path, "test.rtdc")
        ds.export.hdf5(f1, ["temp"], tables=True)

    with new_dataset(f1) as ds2:
        tab = ds2.tables["src_so2exp_cytoshot_monitor"]
        assert tab.has_graphs()
        assert np.allclose(
            tab["focus"][0],
            6.241900073522602,
            atol=1e-7, rtol=0)

        assert tab.meta["COLOR_focus"] == "#a03000"

        tab2 = ds2.tables["src_just_an_array"]
        assert not tab2.has_graphs()
        assert np.array(tab2).shape == (30, 50)
        assert tab2.meta["heinz"] == "kunz"


def test_hdf5_trace_from_tdms():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_2fl-no-image_2017.zip"))

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["trace"])

    ds2 = new_dataset(f1)

    for key in ds["trace"]:
        for ii in range(len(ds)):
            assert np.all(ds["trace"][key][ii] == ds2["trace"][key][ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces():
    """Length of traces is preserved (no filtering)"""
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # sanity check
    assert len(ds) == 7
    assert len(ds["trace"]["fl1_median"]) == 7

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 7
    assert len(ds2["trace"]["fl1_median"]) == 7


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces_filter():
    """Length of traces was wrong when filters were applied #112

    Test dataset length with additional feature.
    """
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # applying some filters
    ds.config["filtering"]["deform min"] = 0.01
    ds.config["filtering"]["deform max"] = 0.1
    ds.apply_filter()

    # sanity check
    assert np.sum(ds.filter.all) == 3

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 3
    assert len(ds2["deform"]) == 3
    assert len(ds2["trace"]["fl1_median"]) == 3
    assert np.all(ds["trace"]["fl1_raw"][3] == ds2["trace"]["fl1_raw"][0])
    assert np.all(ds["trace"]["fl1_raw"][5] == ds2["trace"]["fl1_raw"][1])
    assert np.all(ds["trace"]["fl1_raw"][6] == ds2["trace"]["fl1_raw"][2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_traces_filter2():
    """Length of traces was wrong when filters were applied #112

    Test dataset lenght only with trace.
    """
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))

    # applying some filters
    ds.config["filtering"]["deform min"] = 0.01
    ds.config["filtering"]["deform max"] = 0.1
    ds.apply_filter()

    # sanity check
    assert np.sum(ds.filter.all) == 3

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.rtdc")
    ds.export.hdf5(f1, ["deform", "trace"])

    ds2 = new_dataset(f1)
    assert len(ds2) == 3
    assert len(ds2["deform"]) == 3
    assert len(ds2["trace"]["fl1_median"]) == 3
    assert np.all(ds["trace"]["fl1_raw"][3] == ds2["trace"]["fl1_raw"][0])
    assert np.all(ds["trace"]["fl1_raw"][5] == ds2["trace"]["fl1_raw"][1])
    assert np.all(ds["trace"]["fl1_raw"][6] == ds2["trace"]["fl1_raw"][2])
